#define _USE_MATH_DEFINES

#include "OpenGL.h"

#include "ISceneNode.h"
#include "GroupNode.h"
#include "DrawableSphere.h"
#include "DrawableCube.h"
#include "RenderState.h"
#include "TransformNode.h"
#include "MathUtil.h"
#include "SolidColorMaterial.h"
#include "ShaderProgram.h"
#include "RenderVisitor.h"
#include "LookAtTransform.h"
#include "PerspectiveTransform.h"
#include "MaterialProxy.h"
#include "DrawableProxy.h"
#include "DrawModeNode.h"
#include "DrawableFloor.h"
#include "PrintSceneVisitor.h"
#include "DrawableRenderState.h"
#include "Trackball.h"
#include "SimpleShaderMaterial.h"
#include "LightSetNode.h"
#include "DirectionalLight.h"
#include "PointLight.h"
#include "ShadedMaterial.h"
#include <string>


using namespace OhioState::Graphics;
using namespace OhioState::Math;


ISceneNode* rootSceneNode;
IVisitor* renderVisitor;
PerspectiveTransform* perspectiveNode;
TransformNode* examinerTransform;
const float dollyDelta = 0.1f;
TransformNode* dollyNode;
DrawModeNode* drawMode;
int windowWidth;
int windowHeight;
DrawableRenderState* renderState = new DrawableRenderState("RenderState");
bool showFPS;

enum MouseState { NONE, TRACKBALL, ZOOM, PAN };
MouseState currentMouseState;
OhioState::Controller::Trackball* trackball;
float viewDistance;

void CreateMaterialLibrary()
{
	//
	// This is just a sample. 
	// This should be read in from a file.
	//
	SolidColorMaterial* white = new SolidColorMaterial(Color(1,1,1,1));
	MaterialManager::Instance()->RegisterMaterial("White", white);
	SolidColorMaterial* red = new SolidColorMaterial(Color(1,0,0,1));
	MaterialManager::Instance()->RegisterMaterial("Red", red);
	SolidColorMaterial* blue = new SolidColorMaterial(Color(0,0,1,1));
	MaterialManager::Instance()->RegisterMaterial("Blue", blue);
	SimpleShaderMaterial* whiteShaded = new SimpleShaderMaterial("../Shaders/VertexLight.vert", "../Shaders/SolidColor.frag");
	MaterialManager::Instance()->RegisterMaterial("White Shaded", whiteShaded);
	//SolidColorMaterial* grey = new SolidColorMaterial(Color(0.5f,0.5f,0.5f,1));
	SimpleShaderMaterial* grey = new SimpleShaderMaterial("../Shaders/SolidColor.vert", "../Shaders/SolidColor.frag");
	grey->setColor(Color(0.2f,0.2f,0.2f,1.0f));
	MaterialManager::Instance()->RegisterMaterial("Grey", grey);
	ShadedMaterial* gold = new ShadedMaterial("../Shaders/VertexLight.vert", "../Shaders/SolidColor.frag");
	gold->setAmbientReflection(Color(0.24725, 0.2245, 0.0645, 1.0));
	gold->setDiffuseReflection(Color(0.34615, 0.3143, 0.0903, 1.0));
	gold->setSpecularReflection(Color(0.797357, 0.723991, 0.208006, 1.0));
	gold->setShininess(83.2f);
	MaterialManager::Instance()->RegisterMaterial("Gold", gold);
}

void CreateModelLibrary()
{
	IDrawable* sphere = new DrawableSphere(10000, false);
	ModelManager::Instance()->RegisterModel("Sphere10000", sphere);
	IDrawable* cube = new DrawableCube();
	ModelManager::Instance()->RegisterModel("Cube", cube);
	IDrawable* floor = new DrawableFloor(20,30);
	ModelManager::Instance()->RegisterModel("Floor", floor);
}

void CreateSceneGraph()
{
	//
	// This is just a sample. 
	// This should be read in from a file.
	// Note: bottom up construction!
	//
	DrawableProxy* sphere1 = new DrawableProxy("Sphere1", "Sphere10000");
	DrawableProxy* cube1 = new DrawableProxy("Cube1", "Cube");
	DrawableProxy* floor1 = new DrawableProxy("Floor1", "Floor");
	MaterialProxy* goldSphere = new MaterialProxy("GoldSphere", "Gold");
	MaterialProxy* redCube = new MaterialProxy("RedCube", "Red");
	MaterialProxy* whiteCube = new MaterialProxy("WhiteCube", "White Shaded");
	MaterialProxy* greyFloor = new MaterialProxy("GreyFloor", "Grey");

	GroupNode* floorGroup = new GroupNode("floorGroup");
	floorGroup->AddChild(greyFloor);
	TransformNode* floorTransform = new TransformNode("Floor Transform", floor1);
	floorTransform->Scale(8.0f, 1.0f, 7.0f);
	floorTransform->Translate(0,-0.5f,0);
	floorGroup->AddChild(floorTransform);

	GroupNode* group1 = new GroupNode("group1");
	group1->AddChild(floorGroup);
	TransformNode* sphereTransform = new TransformNode("Sphere Transform", sphere1);
	sphereTransform->Translate(0,1.5f,0);
	group1->AddChild(goldSphere);
	group1->AddChild(sphereTransform);
	group1->AddChild(redCube);
	TransformNode* cubeTransform = new TransformNode("Cube Transform", cube1);
	cubeTransform->Translate(0,0.0f,0);
	cubeTransform->Scale(0.5f,1.5f,0.8f);
	//group1->AddChild(whiteCube);
	group1->AddChild(cubeTransform);

	LightSetNode* lightSet = new LightSetNode("Lights", group1);
	PointLight* light1 = new PointLight();
	light1->setPosition(Vector3(-4,-0.95,0));
	lightSet->AddLight(light1);
	DirectionalLight* light0 = new DirectionalLight(Vector3(1,2,3));
	lightSet->AddLight(light0);

	drawMode = new DrawModeNode("DrawMode", lightSet, DrawModeType::SOLID);

	Vector3 eye(5,0,5);
	Vector3 coi(0,0,0);
	Vector3 viewUp(0,1,0);
	viewDistance = eye.length();
	LookAtTransform* lookAt = new LookAtTransform("Main Camera", drawMode, eye, coi, viewUp);

	TransformNode* preTranslate = new TransformNode("Origin to Center of Rotation", lookAt);
	preTranslate->Translate(0,0,viewDistance);
	examinerTransform = new TransformNode("Examiner Transform", preTranslate);
	TransformNode* postTranslate = new TransformNode("Center of Rotation to Origin", examinerTransform);
	postTranslate->Translate(0,0,-viewDistance);

	dollyNode = new TransformNode("Dolly", postTranslate);

	perspectiveNode = new PerspectiveTransform("Perspective Projection", dollyNode, 45, windowWidth, windowHeight, 0.5, 100);
	
	rootSceneNode = perspectiveNode;
}

void CreateRenderer()
{
	// TODO: create a render-based visitor.
	renderVisitor = new RenderVisitor();
}

void CreateGLUTWindow(std::string windowTitle)
{
	windowWidth = 800;
	windowHeight = 600;
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize( windowWidth, windowHeight );
	glutCreateWindow(windowTitle.c_str());
}

void InitializeOpenGLExtensions()
{
	GLenum err = glewInit();
	if (GLEW_OK != err)
	{
	  /* Problem: glewInit failed, something is seriously wrong. */
	  fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
	  throw "Error initializing GLEW";
	}
	if (!GLEW_VERSION_2_1)
	{
		throw "Fatal Error: OpenGL 2.1 is required";
	}

	fprintf(stdout, "Status: Using GLEW %s\n", glewGetString(GLEW_VERSION));
}

void DisplayScene()
{
	rootSceneNode->Accept(renderVisitor);
}

void DisplayRenderState()
{
	if( showFPS )
		renderState->Draw();
}

void DisplayFrame()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	DisplayScene();
	DisplayRenderState();
	glutSwapBuffers();
}

void ReshapeWindow(int newWidth, int newHeight)
{
	windowWidth = newWidth;
	windowHeight = newHeight;

	perspectiveNode->setWidth(newWidth);
	perspectiveNode->setHeight(newHeight);
}

bool IsQuitting(unsigned char key)
{
	switch(key)
	{
		// Typical quit command keys
		case 'Q':
		case 'q':
		case 'X':
		case 'x':
			return true;
	}
	return false;
}

void ChangeDrawMode(unsigned char key)
{
	//
	// Change the Draw mode if any of the following keys are pressed.
	//
	switch(key)
	{
		case 'w':
			drawMode->setDrawMode(DrawModeType::WIREFRAME);
			::glutPostRedisplay();
			break;
		case 'p':
			drawMode->setDrawMode(DrawModeType::POINTS);
			::glutPostRedisplay();
			break;
		case 's':
			drawMode->setDrawMode(DrawModeType::SOLID);
			::glutPostRedisplay();
			break;
	}
}

void KeyboardController(unsigned char key, int x, int y)
{
	printf("Key Pressed: %c\n", key);
	
	if( IsQuitting(key) )
		exit(0);

	ChangeDrawMode(key);

	//
	// Misc. key controls.
	switch(key)
	{
		case 'z':
			dollyNode->Translate(0,0,dollyDelta);
			::glutPostRedisplay();
			break;
		case 'Z':
			dollyNode->Translate(0,0,-dollyDelta);
			::glutPostRedisplay();
			break;
		case 'f':
			showFPS = !showFPS;
			::glutPostRedisplay();
			break;
	}
}

void NumPadController(int key, int x, int y)
{
	switch(key)
	{
		case GLUT_KEY_LEFT:
			printf("Key Pressed: LEFT ARROW\n");
			examinerTransform->Translate(-0.2f, 0.0f, 0.0f);
			::glutPostRedisplay();
			break;
		case GLUT_KEY_RIGHT:
			printf("Key Pressed: RIGHT ARROW\n");
			examinerTransform->Translate(0.2f, 0.0f, 0.0f);
			::glutPostRedisplay();
			break;
		case GLUT_KEY_UP:
			printf("Key Pressed: UP ARROW\n");
			examinerTransform->Translate(0.0f, -0.2f, 0.0f);
			::glutPostRedisplay();
			break;
		case GLUT_KEY_DOWN:
			printf("Key Pressed: DOWN ARROW\n");
			examinerTransform->Translate(0.0f, 0.2f, 0.0f);
			::glutPostRedisplay();
			break;
	}
}

void MapMouse(int ix, int iy, float& x, float &y)
{
	x = (float) ix / (float) windowWidth; // map to 0 to 1
	x = 2*x - 1; // map to -1 to 1
	y = (float) iy / (float) windowHeight; // map to 0 to 1
	y = 1 - 2*y; // map to 1 to -1 (flip the window coordinates)
}


void MousePressController(int button, int state, int ix, int iy)
{
	if( button == GLUT_LEFT_BUTTON)
	{
		currentMouseState = TRACKBALL;
		float x, y;
		MapMouse(ix,iy,x,y);
		if( state == GLUT_DOWN)
			trackball->BeginManipulation(x, y);
		else
			trackball->EndManipulation(x, y);
	}
}

void MouseMotionController(int ix, int iy)
{
	if( currentMouseState == TRACKBALL )
	{
		float x, y;
		MapMouse(ix,iy,x,y);
		trackball->ManipulateTo(x,y);
		::glutPostRedisplay();
	}
}

void IdleCallback()
{
	examinerTransform->Rotate(0.03f, Vector3(1,0,1));
	::glutPostRedisplay();
}

int main(int argc, char **argv)
{
	glutInit(&argc, argv);
	CreateGLUTWindow("OpenGL Demo Framework");
	InitializeOpenGLExtensions();

	CreateMaterialLibrary();
	CreateModelLibrary();
	CreateSceneGraph();
	CreateRenderer();

	RenderState::UpdateState();

	PrintSceneVisitor* printVisitor = new PrintSceneVisitor();
	rootSceneNode->Accept(printVisitor);

	trackball = new OhioState::Controller::Trackball(examinerTransform);

	glutKeyboardFunc(KeyboardController);
	glutSpecialFunc(NumPadController);
	glutMouseFunc(MousePressController); 
	glutMotionFunc(MouseMotionController); 

	glutDisplayFunc(DisplayFrame);
	glutReshapeFunc(ReshapeWindow);

	//glutIdleFunc(IdleCallback);

	glutMainLoop();
	return 0;
}