#pragma once

#include "OpenGL.h"

#include "IShaderProgram.h"
#include "VertexRoutine.h"
#include "GeometryRoutine.h"
#include "FragmentRoutine.h"
#include <vector>

namespace OhioState
{
	namespace Graphics 
	{
		//
		// Concrete implementation of IShaderProgram.
		// This is an immutable class.
		//
		class ShaderProgram : public IShaderProgram
		{
		public:
			//
			// Constructor. A constructor for a ShaderProgram that uses the old
			// OpenGL Fixed Functional Pipeline model.
			//
			ShaderProgram()
			{
				guid = 0;
				//
				// Set created to true, since we have a fixed-function pipeline.
				created = true;
			}
			//
			// Constructor. A constructor for the simple case of a single vertex 
			// and a single fragment shader (no geometry shader).
			//
			ShaderProgram( VertexRoutine* vertexShader, FragmentRoutine* fragmentShader)
			{
				vertexRoutines.push_back(vertexShader);
				fragmentRoutines.push_back(fragmentShader);

				guid = 0;
				created = false;
			}
			//
			// Constructor. A constructor for the simple case of a single vertex, geometry and 
			// fragment shader.
			//
			ShaderProgram( VertexRoutine* vertexShader, GeometryRoutine* geometryShader, FragmentRoutine* fragmentShader)
			{
				vertexRoutines.push_back(vertexShader);
				geometryRoutines.push_back(geometryShader);
				fragmentRoutines.push_back(fragmentShader);

				created = false;
			}
			//
			// Enable the material.
			//
			virtual void Enable()
			{
				if( !created )
				{
					CreateProgram();
				}
				::glUseProgram( guid );
			}
			//
			// Disable the shader. This will be a no-op for our system, since
			// we will require these to be set at the beginning of every material
			// when it is enabled.
			//
			virtual void Disable()
			{
			}
		private:
			void CreateProgram()
			{
				guid = ::glCreateProgram();
				AttachShaders(vertexRoutines);
				AttachShaders(fragmentRoutines);
				AttachShaders(geometryRoutines);
				Link();
				created = true;
			}

			void AttachShaders(std::vector<IShaderRoutine*> shaders)
			{
				std::vector<IShaderRoutine*>::iterator shaderIterator;
				for( shaderIterator = shaders.begin(); shaderIterator != shaders.end(); shaderIterator++ )
				{
					unsigned int shaderGuid = (*shaderIterator)->getGUID();
					::glAttachShader(guid, shaderGuid);
				}

			}

			void Link()
			{
				::glLinkProgram(guid);
			}

		private:
			std::vector<IShaderRoutine*> vertexRoutines;
			std::vector<IShaderRoutine*> fragmentRoutines;
			std::vector<IShaderRoutine*> geometryRoutines;
			unsigned int guid;
			bool created;
		};
	}
}